@@ -8,10 +8,11 @@ from ..preproc.ccd import FlatField, Log, CCDCorrection
 from ..preproc.shift import VerticalShift
 from ..preproc.double_flatfield import DoubleFlatField
 from ..preproc.phase import PaganinPhaseRetrieval
-from ..preproc.sinogram import SinoProcessing
+from ..preproc.sinogram import SinoProcessing, SinoNormalization
 from ..misc.unsharp import UnsharpMask
+from ..misc.histogram import PartialHistogram, hist_as_2Darray
 from ..resources.utils import is_hdf5_extension
-# ~ from ..io.writer import Writers
+
 
 class FullFieldPipeline:
     """
@@ -29,7 +30,9 @@ class FullFieldPipeline:
     VerticalShiftClass = VerticalShift
     SinoProcessingClass = SinoProcessing
     MLogClass = Log
+    SinoNormalizationClass = SinoNormalization
     FBPClass = None # For now we don't have a plain python/numpy backend for reconstruction
+    HistogramClass = PartialHistogram
 
     def __init__(self, process_config, sub_region, logger=None, extra_options=None, phase_margin=None):
         """
@@ -180,9 +183,13 @@ class FullFieldPipeline:
         return n_recs
 
 
-    def _get_process_name(self):
-        # TODO support "incomplete" processing pipeline
-        return "reconstruction"
+    def _get_process_name(self, kind="reconstruction"):
+        # In the future, might be something like "reconstruction-<ID>"
+        if kind == "reconstruction":
+            return "reconstruction"
+        elif kind == "histogram":
+            return "histogram"
+        return kind
 
 
     def register_callback(self, step_name, callback):
@@ -244,6 +251,8 @@ class FullFieldPipeline:
             shape = self._radios_cropped_shape
         elif step_name == "radios_movements":
             shape = self._radios_cropped_shape
+        elif step_name == "sino_normalization":
+            shape = self._radios_cropped_shape
         elif step_name == "build_sino":
             shape = self._radios_cropped_shape
         elif step_name == "reconstruction":
@@ -307,9 +316,11 @@ class FullFieldPipeline:
         self._init_unsharp()
         self._init_radios_movements()
         self._init_mlog()
+        self._init_sino_normalization()
         self._init_sino_builder()
         self._prepare_reconstruction()
         self._init_reconstruction()
+        self._init_histogram()
         self._init_writer()
 
     @use_options("read_chunk", "chunk_reader")
@@ -347,12 +358,13 @@ class FullFieldPipeline:
     @use_options("double_flatfield", "double_flatfield")
     def _init_double_flatfield(self):
         options = self.processing_options["double_flatfield"]
+        avg_is_on_log = (options["sigma"] is not None)
         self.double_flatfield = self.DoubleFlatFieldClass(
             self._get_shape("double_flatfield"),
             result_url=None,
             input_is_mlog=False,
             output_is_mlog=False,
-            average_is_on_log=False,
+            average_is_on_log=avg_is_on_log,
             sigma_filter=options["sigma"]
         )
 
@@ -384,7 +396,7 @@ class FullFieldPipeline:
             margin=margin,
         )
         if "unsharp_mask" not in self.processing_steps:
-            self.register_callback("phase", self._reshape_radios_after_phase)
+            self.register_callback("phase", FullFieldPipeline._reshape_radios_after_phase)
 
     @use_options("unsharp_mask", "unsharp_mask")
     def _init_unsharp(self):
@@ -394,7 +406,7 @@ class FullFieldPipeline:
             options["unsharp_sigma"], options["unsharp_coeff"],
             mode="reflect", method="gaussian"
         )
-        self.register_callback("unsharp_mask", self._reshape_radios_after_phase)
+        self.register_callback("unsharp_mask", FullFieldPipeline._reshape_radios_after_phase)
 
     @use_options("take_log", "mlog")
     def _init_mlog(self):
@@ -414,6 +426,14 @@ class FullFieldPipeline:
             self._vertical_shifts
         )
 
+    @use_options("sino_normalization", "sino_normalization")
+    def _init_sino_normalization(self):
+        options = self.processing_options["sino_normalization"]
+        self.sino_normalization = self.SinoNormalizationClass(
+            kind=options["method"],
+            radios_shape=self._get_shape("sino_normalization"),
+        )
+
     @use_options("build_sino", "sino_builder")
     def _init_sino_builder(self):
         options = self.processing_options["build_sino"]
@@ -439,6 +459,7 @@ class FullFieldPipeline:
         self._rec_roi = (x_s, x_e, y_s, y_e)
         self._allocate_recs(y_e - y_s, x_e - x_s)
 
+
     @use_options("reconstruction", "reconstruction")
     def _init_reconstruction(self):
         options = self.processing_options["reconstruction"]
@@ -452,6 +473,8 @@ class FullFieldPipeline:
             rot_center = options["rotation_axis_position"]
         if options.get("cor_estimated_auto", False):
             self.logger.info("Estimated center of rotation: %.2f" % rot_center)
+        if self.sino_builder._halftomo_flip:
+            rot_center = self.sino_builder.rot_center
 
         self.reconstruction = self.FBPClass(
             self._get_shape("reconstruction"),
@@ -465,9 +488,16 @@ class FullFieldPipeline:
                 "axis_correction": options["axis_correction"],
             }
         )
-        if options["fbp_filter_type"] == "none":
+        if options["fbp_filter_type"] is None:
             self.reconstruction.fbp = self.reconstruction.backproj
 
+    @use_options("histogram", "histogram")
+    def _init_histogram(self):
+        options = self.processing_options["histogram"]
+        self.histogram = self.HistogramClass(
+            method="fixed_bins_number", num_bins=options["histogram_bins"]
+        )
+
     @use_options("save", "writer")
     def _init_writer(self):
         options = self.processing_options["save"]
@@ -480,7 +510,8 @@ class FullFieldPipeline:
         )
 
         nx_info = None
-        if is_hdf5_extension(options["file_format"]):
+        self._hdf5_output = is_hdf5_extension(options["file_format"])
+        if self._hdf5_output:
             fname_start_index = None
             options["file_prefix"] = self._orig_file_prefix + str(
                 "_%04d" % (self.z_min + self._phase_margin_up)
@@ -574,7 +605,6 @@ class FullFieldPipeline:
     def _take_log(self):
         self.mlog.take_logarithm(self.radios)
 
-
     @pipeline_step("radios_movements", "Applying radios movements")
     def _radios_movements(self, radios=None):
         if radios is None:
@@ -583,6 +613,13 @@ class FullFieldPipeline:
             radios, list(range(radios.shape[0]))
         )
 
+    @pipeline_step("sino_normalization", "Normalizing sinograms")
+    def _normalize_sinos(self, radios=None):
+        if radios is None:
+            radios = self.radios
+        sinos = radios.transpose((1, 0, 2))
+        self.sino_normalization.normalize(sinos)
+
     @pipeline_step("sino_builder", "Building sinograms")
     def _build_sino(self, radios=None):
         if radios is None:
@@ -604,6 +641,12 @@ class FullFieldPipeline:
                 sinos[i], output=self.recs[i]
             )
 
+    @pipeline_step("histogram", "Computing histogram")
+    def _compute_histogram(self, data=None):
+        if data is None:
+            data = self.recs
+        self.recs_histogram = self.histogram.compute_histogram(data.ravel())
+
     @pipeline_step("writer", "Saving data")
     def _write_data(self, data=None):
         if data is None:
@@ -611,6 +654,17 @@ class FullFieldPipeline:
         self.writer.write(data, *self._writer_exec_args, **self._writer_exec_kwargs)
         self.logger.info("Wrote %s" % self.writer.get_filename())
         self.processing_options["save"]["file_prefix"] = self._orig_file_prefix
+        if "histogram" in self.processing_steps and self._hdf5_output:
+            self.logger.info("Saving histogram")
+            self.writer.write(
+                hist_as_2Darray(self.recs_histogram),
+                self._get_process_name(kind="histogram"),
+                processing_index=self._writer_exec_kwargs["processing_index"]+1,
+                config={
+                    "file": path.basename(self.writer.get_filename()),
+                    "bins": self.processing_options["histogram"]["histogram_bins"],
+                }
+            )
 
 
     def _process_chunk(self):
@@ -621,8 +675,10 @@ class FullFieldPipeline:
         self._apply_unsharp()
         self._take_log()
         self._radios_movements()
+        self._normalize_sinos()
         self._build_sino()
         self._reconstruct()
+        self._compute_histogram()
         self._write_data()
         self._process_finalize()
 
