@@ -6,9 +6,10 @@ from ..preproc.shift_cuda import CudaVerticalShift
 from ..preproc.double_flatfield import DoubleFlatField
 from ..preproc.double_flatfield_cuda import CudaDoubleFlatField
 from ..preproc.phase_cuda import CudaPaganinPhaseRetrieval
-from ..preproc.sinogram_cuda import CudaSinoProcessing
-from ..preproc.sinogram import SinoProcessing
+from ..preproc.sinogram_cuda import CudaSinoProcessing, CudaSinoNormalization
+from ..preproc.sinogram import SinoProcessing, SinoNormalization
 from ..misc.unsharp_cuda import CudaUnsharpMask
+from ..misc.histogram import PartialHistogram
 from ..reconstruction.fbp import Backprojector
 from ..cuda.utils import get_cuda_context, __has_pycuda__, __pycuda_error_msg__, copy_big_gpuarray, replace_array_memory
 from .fullfield import FullFieldPipeline
@@ -31,6 +32,8 @@ class CudaFullFieldPipeline(FullFieldPipeline):
     SinoProcessingClass = CudaSinoProcessing
     MLogClass = CudaLog
     FBPClass = Backprojector
+    HistogramClass = PartialHistogram
+    SinoNormalizationClass = CudaSinoNormalization
 
     def __init__(self, process_config, sub_region, logger=None, extra_options=None, phase_margin=None, cuda_options=None):
         self._init_cuda(cuda_options)
@@ -88,11 +91,11 @@ class CudaFullFieldPipeline(FullFieldPipeline):
 
 
     def _register_callbacks(self):
-        self.register_callback("read_chunk", self._read_data_callback)
+        self.register_callback("read_chunk", CudaFullFieldPipeline._read_data_callback)
         if self.reconstruction is not None:
-            self.register_callback("reconstruction", self._rec_callback)
+            self.register_callback("reconstruction", CudaFullFieldPipeline._rec_callback)
         if self.writer is not None:
-            self.register_callback("save", self._saving_callback)
+            self.register_callback("save", CudaFullFieldPipeline._saving_callback)
 
 
 
@@ -113,6 +116,8 @@ class CudaFullFieldPipelineLimitedMemory(CudaFullFieldPipeline):
     DoubleFlatFieldClass = DoubleFlatField
     # VerticalShifts is simpler on host. It should be done on GPU if it slows down the process too much
     VerticalShiftClass = VerticalShift
+    # SinoNormalization is done on host (in-place)
+    SinoNormalizationClass = SinoNormalization
 
     def __init__(self, process_config, sub_region, chunk_size, logger=None, extra_options=None, phase_margin=None, cuda_options=None):
         """
@@ -205,6 +210,8 @@ class CudaFullFieldPipelineLimitedMemory(CudaFullFieldPipeline):
             shape = self._radios_cropped_shape
         elif step_name == "radios_movements":
             shape = self._radios_cropped_shape
+        elif step_name == "sino_normalization":
+            shape = self._radios_cropped_shape
         elif step_name == "build_sino":
             shape = (n_a, self.n_recs, self.radios_shape[-1])
         elif step_name == "reconstruction":
@@ -417,6 +424,9 @@ class CudaFullFieldPipelineLimitedMemory(CudaFullFieldPipeline):
         if self._h_recs is None:
             self._h_recs = np.zeros((self.n_recs, ) + self.recs.shape[1:], "f")
 
+        # Normalize sinos on host
+        self._normalize_sinos(radios=self._h_radios)
+
         # Build sinograms on host.
         # Doing it here rather than in the following loop is simpler but uses more memory.
         # In the loop, "self.sino_builder" has to be re-initialized
@@ -442,6 +452,7 @@ class CudaFullFieldPipelineLimitedMemory(CudaFullFieldPipeline):
             self._reconstruct(sinos=self._d_sinos)
             # Copy D2H
             self._d_recs[:transfer_size].get(ary=self._h_recs[start_idx:end_idx])
+        self._compute_histogram(data=self._h_recs)
         self.logger.debug("End of processing steps on sinos")
         self._h_radios = self._orig_h_radios
         # Write
