@@ -1,19 +1,16 @@
-import numpy as np
-
+import math
 import logging
+import numpy as np
 from numpy.polynomial.polynomial import Polynomial, polyval
-
-from nabu.utils import previouspow2
-from nabu.misc import fourier_filters
+from ..utils import previouspow2
+from ..misc import fourier_filters
 
 try:
     from scipy.ndimage.filters import median_filter
-
     import scipy.fft
 
     local_fftn = scipy.fft.rfftn
     local_ifftn = scipy.fft.irfftn
-
     __have_scipy__ = True
 except ImportError:
     from silx.math.medianfilter import medfilt2d as median_filter
@@ -21,9 +18,7 @@ except ImportError:
 
     local_fftn = np.fft.fftn
     local_ifftn = np.fft.ifftn
-
     __have_scipy__ = False
-
 try:
     import matplotlib.pyplot as plt
     from mpl_toolkits.mplot3d import Axes3D
@@ -31,12 +26,17 @@ try:
     __have_matplotlib__ = True
 except ImportError:
     logging.getLogger(__name__).warning("Matplotlib not available. Plotting disabled")
-
     __have_matplotlib__ = False
 
 
 class AlignmentBase(object):
-    def __init__(self, vert_fft_width=False, horz_fft_width=False, verbose=False, data_type=np.float32):
+    def __init__(
+        self,
+        vert_fft_width=False,
+        horz_fft_width=False,
+        verbose=False,
+        data_type=np.float32,
+    ):
         """
         Alignment basic functions.
 
@@ -86,7 +86,10 @@ class AlignmentBase(object):
             raise ValueError(
                 "The same number of images and positions is required."
                 + " Shape of stack: %s, shape of positions variable: %s"
-                % (" ".join(("%d" % x for x in shape_stack)), " ".join(("%d" % x for x in shape_pos)),)
+                % (
+                    " ".join(("%d" % x for x in shape_stack)),
+                    " ".join(("%d" % x for x in shape_pos)),
+                )
             )
 
     @staticmethod
@@ -94,17 +97,16 @@ class AlignmentBase(object):
         shape_1 = np.squeeze(img_1).shape
         shape_2 = np.squeeze(img_2).shape
         if not len(shape_1) == 2:
-            raise ValueError(
-                "Images need to be 2-dimensional. Shape of image #1: %s" % (" ".join(("%d" % x for x in shape_1)))
-            )
+            raise ValueError("Images need to be 2-dimensional. Shape of image #1: %s" % (" ".join(("%d" % x for x in shape_1))))
         if not len(shape_2) == 2:
-            raise ValueError(
-                "Images need to be 2-dimensional. Shape of image #2: %s" % (" ".join(("%d" % x for x in shape_2)))
-            )
+            raise ValueError("Images need to be 2-dimensional. Shape of image #2: %s" % (" ".join(("%d" % x for x in shape_2))))
         if not np.all(shape_1 == shape_2):
             raise ValueError(
                 "Images need to be of the same shape. Shape of image #1: %s, image #2: %s"
-                % (" ".join(("%d" % x for x in shape_1)), " ".join(("%d" % x for x in shape_2)),)
+                % (
+                    " ".join(("%d" % x for x in shape_1)),
+                    " ".join(("%d" % x for x in shape_2)),
+                )
             )
 
     @staticmethod
@@ -173,7 +175,12 @@ class AlignmentBase(object):
         if np.any(vertex_yx < vertex_min_yx) or np.any(vertex_yx > vertex_max_yx):
             raise ValueError(
                 "Fitted (y: {}, x: {}) positions are outide the input margins y: [{}, {}], and x: [{}, {}]".format(
-                    vertex_yx[0], vertex_yx[1], vertex_min_yx[0], vertex_max_yx[0], vertex_min_yx[1], vertex_max_yx[1],
+                    vertex_yx[0],
+                    vertex_yx[1],
+                    vertex_min_yx[0],
+                    vertex_max_yx[0],
+                    vertex_min_yx[1],
+                    vertex_max_yx[1],
                 )
             )
         return vertex_yx
@@ -239,12 +246,13 @@ class AlignmentBase(object):
         upper_bound_ok = vertex_x < vertex_max_x
         if not np.all(lower_bound_ok * upper_bound_ok):
             if len(f_vals.shape) == 1:
-                message = "Fitted position {} is outide the input margins [{}, {}]".format(
-                    vertex_x, vertex_min_x, vertex_max_x
-                )
+                message = "Fitted position {} is outide the input margins [{}, {}]".format(vertex_x, vertex_min_x, vertex_max_x)
             else:
                 message = "Fitted positions outide the input margins [{}, {}]: %d below and %d above".format(
-                    vertex_min_x, vertex_max_x, np.sum(1 - lower_bound_ok), np.sum(1 - upper_bound_ok),
+                    vertex_min_x,
+                    vertex_max_x,
+                    np.sum(1 - lower_bound_ok),
+                    np.sum(1 - upper_bound_ok),
                 )
             raise ValueError(message)
         if return_vertex_val:
@@ -365,7 +373,13 @@ class AlignmentBase(object):
         return roi_yxhw
 
     def _prepare_image(
-        self, img, invalid_val=1e-5, roi_yxhw=None, median_filt_shape=None, low_pass=None, high_pass=None,
+        self,
+        img,
+        invalid_val=1e-5,
+        roi_yxhw=None,
+        median_filt_shape=None,
+        low_pass=None,
+        high_pass=None,
     ):
         """
         Prepare and returns  a cropped  and filtered image, or array of filtered images if the input is an  array of images.
@@ -393,7 +407,9 @@ class AlignmentBase(object):
 
         if roi_yxhw is not None:
             img = img[
-                ..., roi_yxhw[0] : roi_yxhw[0] + roi_yxhw[2], roi_yxhw[1] : roi_yxhw[1] + roi_yxhw[3],
+                ...,
+                roi_yxhw[0] : roi_yxhw[0] + roi_yxhw[2],
+                roi_yxhw[1] : roi_yxhw[1] + roi_yxhw[3],
             ]
 
         img = img.copy()
@@ -418,7 +434,10 @@ class AlignmentBase(object):
                 # expanding filter shape with ones, to cover the stack of images
                 # but disabling inter-image filtering
                 median_filt_shape = np.concatenate(
-                    (np.ones((len(img_shape) - len(median_filt_shape),), dtype=np.int), median_filt_shape,)
+                    (
+                        np.ones((len(img_shape) - len(median_filt_shape),), dtype=np.int),
+                        median_filt_shape,
+                    )
                 )
                 img = median_filter(img, size=median_filt_shape)
             else:
@@ -559,6 +578,7 @@ class CenterOfRotation(AlignmentBase):
 
         >>> cor_position = CoR_calc.find_shift(radio1, radio2, median_filt_shape=(3, 3))
         """
+
         self._check_img_pair_sizes(img_1, img_2)
 
         if peak_fit_radius < 1:
@@ -574,7 +594,6 @@ class CenterOfRotation(AlignmentBase):
         img_2 = self._prepare_image(img_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
 
         cc = self._compute_correlation_fft(img_1, img_2, padding_mode, high_pass=high_pass, low_pass=low_pass)
-
         img_shape = img_2.shape
         cc_vs = np.fft.fftfreq(img_shape[-2], 1 / img_shape[-2])
         cc_hs = np.fft.fftfreq(img_shape[-1], 1 / img_shape[-1])
@@ -584,6 +603,235 @@ class CenterOfRotation(AlignmentBase):
 
         return fitted_shifts_vh[-1] / 2.0
 
+
+class CenterOfRotationAdaptiveSearch(CenterOfRotation):
+    """This adaptive method works by applying a gaussian which highlights, by apodisation, a region
+    which can possibly contain the good center of rotation.
+    The whole image is spanned during several applications of the apodisation. At each application
+    the apodisation function, which is a gaussian, is moved to a new guess position.
+    The lenght of the step, by which the gaussian is moved, and its sigma are
+    obtained by multiplying the shortest distance from the left or right border with
+    a self.step_fraction and  self.sigma_fraction factors which ensure global overlapping.
+    for each step a region around the CoR  of each image is selected, and the regions of the two images
+    are compared to  calculate a cost function. The value of the cost function, at its minimum
+    is used to select the best step at which the CoR is taken as final result.
+    The option filtered_cost= True (default) triggers the filtering (accordin to low_pass and high_pass)
+    of the two images which are used for he cost function.
+    """
+
+    sigma_fraction = 1.0 / 4.0
+    step_fraction = 1.0 / 6.0
+
+    def find_shift(
+        self,
+        img_1: np.ndarray,
+        img_2: np.ndarray,
+        roi_yxhw=None,
+        median_filt_shape=None,
+        padding_mode=None,
+        peak_fit_radius=1,
+        high_pass=None,
+        low_pass=None,
+        margins=None,
+        filtered_cost=True,
+    ):
+        """Find the Center of Rotation (CoR), given two images.
+
+        This method finds the half-shift between two opposite images, by
+        means of correlation computed in Fourier space.
+        A global search is done on on the detector span (minus a margin) without assuming centered scan conditions.
+
+        The output of this function, allows to compute motor movements for
+        aligning the sample rotation axis. Given the following values:
+
+        - L1: distance from source to motor
+        - L2: distance from source to detector
+        - ps: physical pixel size
+        - v: output of this function
+
+        displacement of motor = (L1 / L2 * ps) * v
+
+        Parameters
+        ----------
+        img_1: numpy.ndarray
+            First image
+        img_2: numpy.ndarray
+            Second image, it needs to have been flipped already (e.g. using numpy.fliplr).
+        roi_yxhw: (2, ) or (4, ) numpy.ndarray, tuple, or array, optional
+            4 elements vector containing: vertical and horizontal coordinates
+            of first pixel, plus height and width of the Region of Interest (RoI).
+            Or a 2 elements vector containing: plus height and width of the
+            centered Region of Interest (RoI).
+            Default is None -> deactivated.
+        median_filt_shape: (2, ) numpy.ndarray, tuple, or array, optional
+            Shape of the median filter window. Default is None -> deactivated.
+        padding_mode: str in numpy.pad's mode list, optional
+            Padding mode, which determines the type of convolution. If None or
+            'wrap' are passed, this resorts to the traditional circular convolution.
+            If 'edge' or 'constant' are passed, it results in a linear convolution.
+            Default is the circular convolution.
+            All options are:
+                None | 'constant' | 'edge' | 'linear_ramp' | 'maximum' | 'mean'
+                | 'median' | 'minimum' | 'reflect' | 'symmetric' |'wrap'
+        peak_fit_radius: int, optional
+            Radius size around the max correlation pixel, for sub-pixel fitting.
+            Minimum and default value is 1.
+        low_pass: float or sequence of two floats.
+            Low-pass filter properties, as described in `nabu.misc.fourier_filters`
+        high_pass: float or sequence of two floats
+            High-pass filter properties, as described in `nabu.misc.fourier_filters`.
+        margins:  None or a couple of floats or ints
+            if margins is None or in the form of  (margin1,margin2) the search is done between margin1 and  dim_x-1-margin2.
+            If left to None then by default (margin1,margin2)  = ( 10, 10 ).
+        filtered_cost: boolean.
+            True by default. It triggers the use of filtered images in the calculation of the cost function.
+
+        Raises
+        ------
+        ValueError
+            In case images are not 2-dimensional or have different sizes.
+
+        Returns
+        -------
+        float
+            Estimated center of rotation position from the center of the RoI in pixels.
+
+        Examples
+        --------
+        The following code computes the center of rotation position for two
+        given images in a tomography scan, where the second image is taken at
+        180 degrees from the first.
+
+        >>> radio1 = data[0, :, :]
+        ... radio2 = np.fliplr(data[1, :, :])
+        ... CoR_calc = CenterOfRotationAdaptiveSearch()
+        ... cor_position = CoR_calc.find_shift(radio1, radio2)
+
+        Or for noisy images:
+
+        >>> cor_position = CoR_calc.find_shift(radio1, radio2, median_filt_shape=(3, 3), high_pass=20, low_pass=1   )
+        """
+
+        if peak_fit_radius < 1:
+            logging.getLogger(__name__).warning(
+                "Parameter peak_fit_radius should be at least 1, given: %d instead." % peak_fit_radius
+            )
+            peak_fit_radius = 1
+
+        self._check_img_pair_sizes(img_1, img_2)
+
+        used_type = img_1.dtype
+
+        roi_yxhw = self._determine_roi(img_1.shape, roi_yxhw)
+
+        if filtered_cost and (low_pass is not None or high_pass is not None):
+            img_filter = fourier_filters.get_bandpass_filter(
+                img_1.shape[-2:],
+                cutoff_lowpass=low_pass,
+                cutoff_highpass=high_pass,
+                use_rfft=__have_scipy__,
+                data_type=self.data_type,
+            )
+            # fft2 and iff2 use axes=(-2, -1) by default
+            img_filtered_1 = local_ifftn(local_fftn(img_1, axes=(-2, -1)) * img_filter, axes=(-2, -1)).real
+            img_filtered_2 = local_ifftn(local_fftn(img_2, axes=(-2, -1)) * img_filter, axes=(-2, -1)).real
+        else:
+            img_filtered_1 = img_1
+            img_filtered_2 = img_2
+
+        img_1 = self._prepare_image(img_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
+        img_2 = self._prepare_image(img_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
+
+        img_filtered_1 = self._prepare_image(img_filtered_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
+        img_filtered_2 = self._prepare_image(img_filtered_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
+
+        dim_radio = img_1.shape[1]
+
+        if margins is None:
+            lim_1, lim_2 = 10, dim_radio - 1 - 10
+        else:
+            lim_1, lim_2 = margins
+            lim_2 = dim_radio - 1 - lim_2
+
+        if lim_1 < 1:
+            lim_1 = 1
+        if lim_2 > dim_radio - 2:
+            lim_2 = dim_radio - 2
+
+        if lim_2 <= lim_1:
+            message = (
+                "Image shape or cropped selection too small for global search. After removal of the margins the search limits collide. The cropped size is %d\n"
+                % (dim_radio)
+            )
+            raise ValueError(message)
+
+        found_centers = []
+        x_cor = lim_1
+        while x_cor < lim_2:
+
+            tmp_sigma = (
+                min(
+                    (img_1.shape[1] - x_cor),
+                    (x_cor),
+                )
+                * self.sigma_fraction
+            )
+
+            tmp_x = (np.arange(img_1.shape[1]) - x_cor) / tmp_sigma
+            apodis = np.exp(-tmp_x * tmp_x / 2.0)
+
+            x_cor_rel = x_cor - (img_1.shape[1] // 2)
+
+            img_1_apodised = img_1 * apodis
+
+            cor_position = super(self.__class__, self).find_shift(
+                img_1_apodised.astype(used_type),
+                img_2.astype(used_type),
+                low_pass=low_pass,
+                high_pass=high_pass,
+                roi_yxhw=roi_yxhw,
+            )
+
+            p_1 = cor_position * 2
+            if cor_position < 0:
+                p_2 = img_2.shape[1] + cor_position * 2
+            else:
+                p_2 = -img_2.shape[1] + cor_position * 2
+
+            if abs(x_cor_rel - p_1 / 2) < abs(x_cor_rel - p_2 / 2):
+                cor_position = p_1 / 2
+            else:
+                cor_position = p_2 / 2
+
+            cor_in_img = img_1.shape[1] // 2 + cor_position
+            tmp_sigma = (
+                min(
+                    (img_1.shape[1] - cor_in_img),
+                    (cor_in_img),
+                )
+                * self.sigma_fraction
+            )
+
+            M1 = int(round(cor_position + img_1.shape[1] // 2)) - int(round(tmp_sigma))
+            M2 = int(round(cor_position + img_1.shape[1] // 2)) + int(round(tmp_sigma))
+
+            piece_1 = img_filtered_1[:, M1:M2]
+            piece_2 = img_filtered_2[:, img_1.shape[1] - M2 : img_1.shape[1] - M1]
+            piece_1 = piece_1 - piece_1.mean()
+            piece_2 = piece_2 - piece_2.mean()
+            energy = np.array(piece_1 * piece_1 + piece_2 * piece_2, "d").sum()
+            diff_energy = np.array((piece_1 - piece_2) * (piece_1 - piece_2), "d").sum()
+            cost = diff_energy / energy
+
+            if not np.isnan(cost):
+                found_centers.append([cost, abs(x_cor_rel - cor_position), cor_position, energy])
+
+            x_cor = min(x_cor + x_cor * self.step_fraction, x_cor + (dim_radio - x_cor) * self.step_fraction)
+
+        found_centers.sort()
+        cor_position = found_centers[0][2]
+        return cor_position
+
     __call__ = find_shift
 
 
@@ -758,10 +1006,10 @@ class DetectorTranslationAlongBeam(AlignmentBase):
             print("Fitted pixel shifts per unit-length: vertical = %f, horizontal = %f" % (coeffs_v[1], coeffs_h[1]))
             f, axs = plt.subplots(1, 2)
             axs[0].scatter(img_pos, shifts_vh[:, 0])
-            axs[0].plot(img_pos, polyval(img_pos, coeffs_v), '-C1')
+            axs[0].plot(img_pos, polyval(img_pos, coeffs_v), "-C1")
             axs[0].set_title("Vertical shifts")
             axs[1].scatter(img_pos, shifts_vh[:, 1])
-            axs[1].plot(img_pos, polyval(img_pos, coeffs_h), '-C1')
+            axs[1].plot(img_pos, polyval(img_pos, coeffs_h), "-C1")
             axs[1].set_title("Horizontal shifts")
             plt.show(block=False)
 
@@ -863,7 +1111,14 @@ class CameraTilt(CenterOfRotation):
         img_1 = self._prepare_image(img_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
         img_2 = self._prepare_image(img_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
 
-        cc = self._compute_correlation_fft(img_1, img_2, padding_mode, axes=(-1,), high_pass=high_pass, low_pass=low_pass,)
+        cc = self._compute_correlation_fft(
+            img_1,
+            img_2,
+            padding_mode,
+            axes=(-1,),
+            high_pass=high_pass,
+            low_pass=low_pass,
+        )
 
         img_shape = img_2.shape
         cc_h_coords = np.fft.fftfreq(img_shape[-1], 1 / img_shape[-1])
@@ -876,7 +1131,7 @@ class CameraTilt(CenterOfRotation):
         if __have_scipy__:
             fitted_shifts_h = median_filter(fitted_shifts_h, size=3)
         else:
-            fitted_shifts_h = median_filter_1d(fitted_shifts_h, size=3)
+            fitted_shifts_h = median_filter_1d(fitted_shifts_h, kernel_size=3)
 
         half_img_size = (img_shape[-2] - 1) / 2
         cc_v_coords = np.linspace(-half_img_size, half_img_size, img_shape[-2])
@@ -887,11 +1142,14 @@ class CameraTilt(CenterOfRotation):
 
         if self.verbose:
             print(
-                "Fitted center of rotation (pixels):", cor_offset_pix, "and camera tilt (degrees):", tilt_deg,
+                "Fitted center of rotation (pixels):",
+                cor_offset_pix,
+                "and camera tilt (degrees):",
+                tilt_deg,
             )
             f, ax = plt.subplots(1, 1)
             ax.scatter(cc_v_coords, fitted_shifts_h)
-            ax.plot(cc_v_coords, polyval(cc_v_coords, coeffs_h), '-C1')
+            ax.plot(cc_v_coords, polyval(cc_v_coords, coeffs_h), "-C1")
             ax.set_title("Correlation peaks")
             plt.show(block=False)
 
@@ -982,7 +1240,11 @@ class CameraFocus(CenterOfRotation):
         roi_yxhw = self._determine_roi(img_shape, roi_yxhw)
 
         img_stack = self._prepare_image(
-            img_stack, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape, low_pass=low_pass, high_pass=high_pass,
+            img_stack,
+            roi_yxhw=roi_yxhw,
+            median_filt_shape=median_filt_shape,
+            low_pass=low_pass,
+            high_pass=high_pass,
         )
 
         img_stds = np.std(img_stack, axis=(-2, -1)) / np.mean(img_stack, axis=(-2, -1))
@@ -999,10 +1261,15 @@ class CameraFocus(CenterOfRotation):
         focus_ind += 1
 
         if self.verbose:
-            print("Fitted focus motor position:", focus_pos, "and corresponding image position:", focus_ind)
+            print(
+                "Fitted focus motor position:",
+                focus_pos,
+                "and corresponding image position:",
+                focus_ind,
+            )
             f, ax = plt.subplots(1, 1)
             ax.stem(img_pos, img_stds)
-            ax.stem(focus_pos, img_std_max, linefmt='C1-', markerfmt='C1o')
+            ax.stem(focus_pos, img_std_max, linefmt="C1-", markerfmt="C1o")
             ax.set_title("Images std")
             plt.show(block=False)
 
@@ -1179,14 +1446,19 @@ class CameraFocus(CenterOfRotation):
             self._check_img_block_size(roi_yxhw[2:], regions_number, suggest_new_shape=False)
 
         img_stack = self._prepare_image(
-            img_stack, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape, low_pass=low_pass, high_pass=high_pass,
+            img_stack,
+            roi_yxhw=roi_yxhw,
+            median_filt_shape=median_filt_shape,
+            low_pass=low_pass,
+            high_pass=high_pass,
         )
         img_shape = img_stack.shape[-2:]
 
         block_size = np.array(img_shape) / regions_number
-        block_stack_size = np.array([
-            num_imgs, regions_number, block_size[-2], regions_number, block_size[-1]
-        ], dtype=np.int)
+        block_stack_size = np.array(
+            [num_imgs, regions_number, block_size[-2], regions_number, block_size[-1]],
+            dtype=np.int,
+        )
         img_stack = np.reshape(img_stack, block_stack_size)
 
         img_stds = np.std(img_stack, axis=(-3, -1)) / np.mean(img_stack, axis=(-3, -1))
@@ -1213,16 +1485,31 @@ class CameraFocus(CenterOfRotation):
         focus_ind = np.mean(focus_inds) + 1
 
         if self.verbose:
-            print("Fitted focus motor position:", focus_pos, "and corresponding image position:", focus_ind)
+            print(
+                "Fitted focus motor position:",
+                focus_pos,
+                "and corresponding image position:",
+                focus_ind,
+            )
             print("Fitted tilts (to be divided by pixel size, and converted to deg): (v, h) %s" % tilts_vh)
             fig = plt.figure()
-            ax = fig.add_subplot(111, projection='3d')
+            ax = fig.add_subplot(111, projection="3d")
             ax.plot_wireframe(fx, fy, focus_poss)
             regions_half_shape = (regions_number - 1) / 2
             base_points = np.linspace(-regions_half_shape, regions_half_shape, regions_number)
-            ax.plot(np.zeros((regions_number, )), base_points, np.polyval([tg_v, focus_pos], base_points), 'C2')
-            ax.plot(base_points, np.zeros((regions_number, )), np.polyval([tg_h, focus_pos], base_points), 'C2')
-            ax.scatter(0, 0, focus_pos, marker='o', c='C1')
+            ax.plot(
+                np.zeros((regions_number,)),
+                base_points,
+                np.polyval([tg_v, focus_pos], base_points),
+                "C2",
+            )
+            ax.plot(
+                base_points,
+                np.zeros((regions_number,)),
+                np.polyval([tg_h, focus_pos], base_points),
+                "C2",
+            )
+            ax.scatter(0, 0, focus_pos, marker="o", c="C1")
             ax.set_title("Images std")
             plt.show(block=False)
 
