@@ -34,10 +34,15 @@ def bootstrap():
             exit(0)
 
     if do_bootstrap:
+        prefilled_values = {}
+        if args["dataset"] != "":
+            prefilled_values["dataset"] = {}
+            prefilled_values["dataset"]["location"] = args["dataset"]
         generate_nabu_configfile(
             args["output"],
             comments=not(no_comments),
             options_level=args["level"],
+            prefilled_values=prefilled_values,
         )
     else: # do_convert
         print("PyHST file converter is not implemented yet")
